"""Provide constants used throughout the package."""

MINUTE = 60
DAY = 24 * 60 * MINUTE

#: The number of Joules in a Watt-hour.
wh_to_j = 3600
#: The number of Joules in a kiloWatt-hour.
kwh_to_j = 1000 * wh_to_j
